"""
tags for common CI attributes
"""
import os
import re

from . import git


# Stage Name
STAGE_NAME = "ci.stage.name"

# Job Name
JOB_NAME = "ci.job.name"

# Job URL
JOB_URL = "ci.job.url"

# Pipeline ID
PIPELINE_ID = "ci.pipeline.id"

# Pipeline Name
PIPELINE_NAME = "ci.pipeline.name"

# Pipeline Number
PIPELINE_NUMBER = "ci.pipeline.number"

# Pipeline URL
PIPELINE_URL = "ci.pipeline.url"

# Provider
PROVIDER_NAME = "ci.provider.name"

# Workspace Path
WORKSPACE_PATH = "ci.workspace_path"

_RE_REFS = re.compile(r"^refs/(heads/)?")
_RE_ORIGIN = re.compile(r"^origin/")
_RE_TAGS = re.compile(r"^tags/")
_RE_URL = re.compile(r"(https?://)[^/]*@")


def _normalize_ref(name):
    return _RE_TAGS.sub("", _RE_ORIGIN.sub("", _RE_REFS.sub("", name))) if name is not None else None


def _filter_sensitive_info(url):
    return _RE_URL.sub("\\1", url) if url is not None else None


def tags(env=None):
    env = os.environ if env is None else env
    tags = {}
    for key, extract in PROVIDERS:
        if key in env:
            tags = extract(env)
            break

    tags[git.TAG] = _normalize_ref(tags.get(git.TAG))
    if tags.get(git.TAG) and git.BRANCH in tags:
        del tags[git.BRANCH]
    tags[git.BRANCH] = _normalize_ref(tags.get(git.BRANCH))
    tags[git.REPOSITORY_URL] = _filter_sensitive_info(tags.get(git.REPOSITORY_URL))

    # Expand ~
    workspace_path = tags.get(WORKSPACE_PATH)
    if workspace_path:
        tags[WORKSPACE_PATH] = os.path.expanduser(workspace_path)

    return {k: v for k, v in tags.items() if v is not None}


def extract_appveyor(env):
    url = "https://ci.appveyor.com/project/{0}/builds/{1}".format(
        env.get("APPVEYOR_REPO_NAME"), env.get("APPVEYOR_BUILD_ID")
    )

    if env.get("APPVEYOR_REPO_PROVIDER") == "github":
        repository = "https://github.com/{0}.git".format(env.get("APPVEYOR_REPO_NAME"))
        commit = env.get("APPVEYOR_REPO_COMMIT")
        branch = env.get("APPVEYOR_PULL_REQUEST_HEAD_REPO_BRANCH") or env.get("APPVEYOR_REPO_BRANCH")
        tag = env.get("APPVEYOR_REPO_TAG_NAME")
    else:
        repository = commit = branch = tag = None

    return {
        PROVIDER_NAME: "appveyor",
        git.REPOSITORY_URL: repository,
        git.COMMIT_SHA: commit,
        WORKSPACE_PATH: env.get("APPVEYOR_BUILD_FOLDER"),
        PIPELINE_ID: env.get("APPVEYOR_BUILD_ID"),
        PIPELINE_NAME: env.get("APPVEYOR_REPO_NAME"),
        PIPELINE_NUMBER: env.get("APPVEYOR_BUILD_NUMBER"),
        PIPELINE_URL: url,
        JOB_URL: url,
        git.BRANCH: branch,
        git.TAG: tag,
    }


def extract_azure_pipelines(env):
    if env.get("SYSTEM_TEAMFOUNDATIONSERVERURI") and env.get("SYSTEM_TEAMPROJECTID") and env.get("BUILD_BUILDID"):
        base_url = "{0}{1}/_build/results?buildId={2}".format(
            env.get("SYSTEM_TEAMFOUNDATIONSERVERURI"), env.get("SYSTEM_TEAMPROJECTID"), env.get("BUILD_BUILDID")
        )
        pipeline_url = base_url
        job_url = base_url + "&view=logs&j={0}&t={1}".format(env.get("SYSTEM_JOBID"), env.get("SYSTEM_TASKINSTANCEID"))
    else:
        pipeline_url = job_url = None
    branch_or_tag = (
        env.get("SYSTEM_PULLREQUEST_SOURCEBRANCH") or env.get("BUILD_SOURCEBRANCH") or env.get("BUILD_SOURCEBRANCHNAME")
    )
    return {
        PROVIDER_NAME: "azurepipelines",
        WORKSPACE_PATH: env.get("BUILD_SOURCESDIRECTORY"),
        PIPELINE_ID: env.get("BUILD_BUILDID"),
        PIPELINE_NAME: env.get("BUILD_DEFINITIONNAME"),
        PIPELINE_NUMBER: env.get("BUILD_BUILDID"),
        PIPELINE_URL: pipeline_url,
        JOB_URL: job_url,
        git.REPOSITORY_URL: env.get("SYSTEM_PULLREQUEST_SOURCEREPOSITORYURI") or env.get("BUILD_REPOSITORY_URI"),
        git.COMMIT_SHA: env.get("SYSTEM_PULLREQUEST_SOURCECOMMITID") or env.get("BUILD_SOURCEVERSION"),
        git.BRANCH: branch_or_tag if "tags/" not in branch_or_tag else None,
        git.TAG: branch_or_tag if "tags/" in branch_or_tag else None,
    }


def extract_bitbucket(env):
    url = "https://bitbucket.org/{0}/addon/pipelines/home#!/results/{1}".format(
        env.get("BITBUCKET_REPO_FULL_NAME"), env.get("BITBUCKET_BUILD_NUMBER")
    )
    return {
        git.BRANCH: env.get("BITBUCKET_BRANCH"),
        git.COMMIT_SHA: env.get("BITBUCKET_COMMIT"),
        git.REPOSITORY_URL: env.get("BITBUCKET_GIT_SSH_ORIGIN"),
        git.TAG: env.get("BITBUCKET_TAG"),
        JOB_URL: url,
        PIPELINE_ID: env.get("BITBUCKET_PIPELINE_UUID", "").strip("{}}") or None,
        PIPELINE_NAME: env.get("BITBUCKET_REPO_FULL_NAME"),
        PIPELINE_NUMBER: env.get("BITBUCKET_BUILD_NUMBER"),
        PIPELINE_URL: url,
        PROVIDER_NAME: "bitbucket",
        WORKSPACE_PATH: env.get("BITBUCKET_CLONE_DIR"),
    }


def extract_buildkite(env):
    return {
        git.BRANCH: env.get("BUILDKITE_BRANCH"),
        git.COMMIT_SHA: env.get("BUILDKITE_COMMIT"),
        git.REPOSITORY_URL: env.get("BUILDKITE_REPO"),
        git.TAG: env.get("BUILDKITE_TAG"),
        PIPELINE_ID: env.get("BUILDKITE_BUILD_ID"),
        PIPELINE_NAME: env.get("BUILDKITE_PIPELINE_SLUG"),
        PIPELINE_NUMBER: env.get("BUILDKITE_BUILD_NUMBER"),
        PIPELINE_URL: env.get("BUILDKITE_BUILD_URL"),
        JOB_URL: "{0}#{1}".format(env.get("BUILDKITE_BUILD_URL"), env.get("BUILDKITE_JOB_ID")),
        PROVIDER_NAME: "buildkite",
        WORKSPACE_PATH: env.get("BUILDKITE_BUILD_CHECKOUT_PATH"),
    }


def extract_circle_ci(env):
    return {
        git.BRANCH: env.get("CIRCLE_BRANCH"),
        git.COMMIT_SHA: env.get("CIRCLE_SHA1"),
        git.REPOSITORY_URL: env.get("CIRCLE_REPOSITORY_URL"),
        git.TAG: env.get("CIRCLE_TAG"),
        PIPELINE_ID: env.get("CIRCLE_WORKFLOW_ID"),
        PIPELINE_NAME: env.get("CIRCLE_PROJECT_REPONAME"),
        PIPELINE_NUMBER: env.get("CIRCLE_BUILD_NUM"),
        PIPELINE_URL: env.get("CIRCLE_BUILD_URL"),
        JOB_URL: env.get("CIRCLE_BUILD_URL"),
        PROVIDER_NAME: "circleci",
        WORKSPACE_PATH: env.get("CIRCLE_WORKING_DIRECTORY"),
    }


def extract_github_actions(env):
    branch_or_tag = env.get("GITHUB_HEAD_REF") or env.get("GITHUB_REF")
    branch = tag = None
    if "tags/" in branch_or_tag:
        tag = branch_or_tag
    else:
        branch = branch_or_tag
    return {
        git.BRANCH: branch,
        git.COMMIT_SHA: env.get("GITHUB_SHA"),
        git.REPOSITORY_URL: "https://github.com/{0}.git".format(env.get("GITHUB_REPOSITORY")),
        git.TAG: tag,
        JOB_URL: "https://github.com/{0}/commit/{1}/checks".format(env.get("GITHUB_REPOSITORY"), env.get("GITHUB_SHA")),
        PIPELINE_ID: env.get("GITHUB_RUN_ID"),
        PIPELINE_NAME: env.get("GITHUB_WORKFLOW"),
        PIPELINE_NUMBER: env.get("GITHUB_RUN_NUMBER"),
        PIPELINE_URL: "https://github.com/{0}/commit/{1}/checks".format(
            env.get("GITHUB_REPOSITORY"), env.get("GITHUB_SHA")
        ),
        PROVIDER_NAME: "github",
        WORKSPACE_PATH: env.get("GITHUB_WORKSPACE"),
    }


def extract_gitlab(env):
    url = env.get("CI_PIPELINE_URL")
    if url:
        url = re.sub("/-/pipelines/", "/pipelines/", url)
    return {
        git.BRANCH: env.get("CI_COMMIT_BRANCH"),
        git.COMMIT_SHA: env.get("CI_COMMIT_SHA"),
        git.REPOSITORY_URL: env.get("CI_REPOSITORY_URL"),
        git.TAG: env.get("CI_COMMIT_TAG"),
        STAGE_NAME: env.get("CI_JOB_STAGE"),
        JOB_NAME: env.get("CI_JOB_NAME"),
        JOB_URL: env.get("CI_JOB_URL"),
        PIPELINE_ID: env.get("CI_PIPELINE_ID"),
        PIPELINE_NAME: env.get("CI_PROJECT_PATH"),
        PIPELINE_NUMBER: env.get("CI_PIPELINE_IID"),
        PIPELINE_URL: url,
        PROVIDER_NAME: "gitlab",
        WORKSPACE_PATH: env.get("CI_PROJECT_DIR"),
    }


def extract_jenkins(env):
    branch_or_tag = env.get("GIT_BRANCH")
    branch = tag = None
    if "tags/" in branch_or_tag:
        tag = branch_or_tag
    else:
        branch = branch_or_tag
    name = env.get("JOB_NAME")
    if name and branch:
        name = re.sub("/" + _normalize_ref(branch), "", name)
    if name:
        name = "/".join((v for v in name.split("/") if v and "=" not in v))

    return {
        git.BRANCH: branch,
        git.COMMIT_SHA: env.get("GIT_COMMIT"),
        git.REPOSITORY_URL: env.get("GIT_URL"),
        git.TAG: tag,
        PIPELINE_ID: env.get("BUILD_TAG"),
        PIPELINE_NAME: name,
        PIPELINE_NUMBER: env.get("BUILD_NUMBER"),
        PIPELINE_URL: env.get("BUILD_URL"),
        PROVIDER_NAME: "jenkins",
        WORKSPACE_PATH: env.get("WORKSPACE"),
    }


def extract_teamcity(env):
    return {
        git.COMMIT_SHA: env.get("BUILD_VCS_NUMBER"),
        git.REPOSITORY_URL: env.get("BUILD_VCS_URL"),
        PIPELINE_ID: env.get("BUILD_ID"),
        PIPELINE_NUMBER: env.get("BUILD_NUMBER"),
        PIPELINE_URL: (
            "{0}/viewLog.html?buildId={1}".format(env.get("SERVER_URL"), env.get("BUILD_ID"))
            if env.get("SERVER_URL") and env.get("BUILD_ID")
            else None
        ),
        PROVIDER_NAME: "teamcity",
        WORKSPACE_PATH: env.get("BUILD_CHECKOUTDIR"),
    }


def extract_travis(env):
    return {
        git.BRANCH: env.get("TRAVIS_PULL_REQUEST_BRANCH") or env.get("TRAVIS_BRANCH"),
        git.COMMIT_SHA: env.get("TRAVIS_COMMIT"),
        git.REPOSITORY_URL: "https://github.com/{0}.git".format(env.get("TRAVIS_REPO_SLUG")),
        git.TAG: env.get("TRAVIS_TAG"),
        JOB_URL: env.get("TRAVIS_JOB_WEB_URL"),
        PIPELINE_ID: env.get("TRAVIS_BUILD_ID"),
        PIPELINE_NAME: env.get("TRAVIS_REPO_SLUG"),
        PIPELINE_NUMBER: env.get("TRAVIS_BUILD_NUMBER"),
        PIPELINE_URL: env.get("TRAVIS_BUILD_WEB_URL"),
        PROVIDER_NAME: "travisci",
        WORKSPACE_PATH: env.get("TRAVIS_BUILD_DIR"),
    }


def extract_bitrise(env):
    commit = env.get("BITRISE_GIT_COMMIT") or env.get("GIT_CLONE_COMMIT_HASH")
    branch = env.get("BITRISEIO_GIT_BRANCH_DEST") or env.get("BITRISE_GIT_BRANCH")
    return {
        PROVIDER_NAME: "bitrise",
        PIPELINE_ID: env.get("BITRISE_BUILD_SLUG"),
        PIPELINE_NAME: env.get("BITRISE_APP_TITLE"),
        PIPELINE_NUMBER: env.get("BITRISE_BUILD_NUMBER"),
        PIPELINE_URL: env.get("BITRISE_BUILD_URL"),
        WORKSPACE_PATH: env.get("BITRISE_SOURCE_DIR"),
        git.REPOSITORY_URL: env.get("GIT_REPOSITORY_URL"),
        git.COMMIT_SHA: commit,
        git.BRANCH: branch,
        git.TAG: env.get("BITRISE_GIT_TAG"),
    }


PROVIDERS = (
    ("APPVEYOR", extract_appveyor),
    ("TF_BUILD", extract_azure_pipelines),
    ("BITBUCKET_COMMIT", extract_bitbucket),
    ("BUILDKITE", extract_buildkite),
    ("CIRCLECI", extract_circle_ci),
    ("GITHUB_SHA", extract_github_actions),
    ("GITLAB_CI", extract_gitlab),
    ("JENKINS_URL", extract_jenkins),
    ("TEAMCITY_VERSION", extract_teamcity),
    ("TRAVIS", extract_travis),
    ("BITRISE_BUILD_SLUG", extract_bitrise),
)



import mysql.connector

from ddtrace import Pin
from ddtrace import config
from ddtrace.contrib.dbapi import TracedConnection
from ddtrace.vendor import wrapt

from ...ext import db
from ...ext import net
from ...utils.formats import asbool
from ...utils.formats import get_env


config._add(
    "mysql",
    dict(
        _default_service="mysql",
        trace_fetch_methods=asbool(get_env("mysql", "trace_fetch_methods", default=False)),
        _deprecated_name="dbapi2",
    ),
)

CONN_ATTR_BY_TAG = {
    net.TARGET_HOST: "server_host",
    net.TARGET_PORT: "server_port",
    db.USER: "user",
    db.NAME: "database",
}


def patch():
    wrapt.wrap_function_wrapper("mysql.connector", "connect", _connect)
    # `Connect` is an alias for `connect`, patch it too
    if hasattr(mysql.connector, "Connect"):
        mysql.connector.Connect = mysql.connector.connect


def unpatch():
    if isinstance(mysql.connector.connect, wrapt.ObjectProxy):
        mysql.connector.connect = mysql.connector.connect.__wrapped__
        if hasattr(mysql.connector, "Connect"):
            mysql.connector.Connect = mysql.connector.connect


def _connect(func, instance, args, kwargs):
    conn = func(*args, **kwargs)
    return patch_conn(conn)


def patch_conn(conn):

    tags = {t: getattr(conn, a) for t, a in CONN_ATTR_BY_TAG.items() if getattr(conn, a, "") != ""}
    pin = Pin(app="mysql", tags=tags)

    # grab the metadata from the conn
    wrapped = TracedConnection(conn, pin=pin, cfg=config.mysql)
    pin.onto(wrapped)
    return wrapped



import abc

import six

from .context import Context
from .internal.compat import contextvars


_DD_CONTEXTVAR = contextvars.ContextVar("datadog_contextvar", default=None)


class BaseContextProvider(six.with_metaclass(abc.ABCMeta)):
    """
    A ``ContextProvider`` is an interface that provides the blueprint
    for a callable class, capable to retrieve the current active
    ``Context`` instance. Context providers must inherit this class
    and implement:
    * the ``active`` method, that returns the current active ``Context``
    * the ``activate`` method, that sets the current active ``Context``
    """

    @abc.abstractmethod
    def _has_active_context(self):
        pass

    @abc.abstractmethod
    def activate(self, context):
        # type: (Context) -> None
        pass

    @abc.abstractmethod
    def active(self):
        # type: () -> Context
        pass

    def __call__(self, *args, **kwargs):
        """Method available for backward-compatibility. It proxies the call to
        ``self.active()`` and must not do anything more.
        """
        return self.active()


class DefaultContextProvider(BaseContextProvider):
    """
    Default context provider that retrieves all contexts from the current
    thread-local storage. It is suitable for synchronous programming and
    Python WSGI frameworks.
    """

    def __init__(self):
        # type: () -> None
        _DD_CONTEXTVAR.set(None)

    def _has_active_context(self):
        """
        Check whether we have a currently active context.

        :returns: Whether we have an active context
        :rtype: bool
        """
        ctx = _DD_CONTEXTVAR.get()
        return ctx is not None

    def activate(self, ctx):
        # type: (Context) -> None
        """Makes the given ``context`` active, so that the provider calls
        the thread-local storage implementation.
        """
        _DD_CONTEXTVAR.set(ctx)  # type: ignore[arg-type]

    def active(self):
        # type: () -> Context
        """Returns the current active ``Context`` for this tracer. Returned
        ``Context`` must be thread-safe or thread-local for this specific
        implementation.
        """
        ctx = _DD_CONTEXTVAR.get()
        if not ctx:
            ctx = Context()
            self.activate(ctx)

        return ctx



import os
import sys

import pytest

from ddtrace import Pin
from ddtrace.ext import test
from tests.utils import TracerTestCase


class TestPytest(TracerTestCase):
    @pytest.fixture(autouse=True)
    def fixtures(self, testdir, monkeypatch):
        self.testdir = testdir
        self.monkeypatch = monkeypatch

    def inline_run(self, *args):
        """Execute test script with test tracer."""

        class PinTracer:
            @staticmethod
            def pytest_configure(config):
                if Pin.get_from(config) is not None:
                    Pin.override(config, tracer=self.tracer)

        return self.testdir.inline_run(*args, plugins=[PinTracer()])

    def subprocess_run(self, *args):
        """Execute test script with test tracer."""
        return self.testdir.runpytest_subprocess(*args)

    @pytest.mark.skipif(sys.version_info[0] == 2, reason="Triggers a bug with coverage, sqlite and Python 2")
    def test_patch_all(self):
        """Test with --ddtrace-patch-all."""
        py_file = self.testdir.makepyfile(
            """
            import ddtrace.monkey

            def test_patched_all():
                assert ddtrace.monkey._PATCHED_MODULES
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace-patch-all", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    @pytest.mark.skipif(sys.version_info[0] == 2, reason="Triggers a bug with coverage, sqlite and Python 2")
    def test_patch_all_init(self):
        """Test with ddtrace-patch-all via ini."""
        self.testdir.makefile(".ini", pytest="[pytest]\nddtrace-patch-all=1\n")
        py_file = self.testdir.makepyfile(
            """
            import ddtrace.monkey

            def test_patched_all():
                assert ddtrace.monkey._PATCHED_MODULES
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    def test_disabled(self):
        """Test without --ddtrace."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            def test_no_trace(ddspan):
                assert ddspan is None
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 0

    def test_ini(self):
        """Test ini config."""
        self.testdir.makefile(".ini", pytest="[pytest]\nddtrace=1\n")
        py_file = self.testdir.makepyfile(
            """
            import pytest

            def test_ini(ddspan):
                assert ddspan is not None
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run(file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1

    def test_parameterize_case(self):
        """Test parametrize case."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.parametrize('abc', [1, 2, 3, 4, pytest.param(5, marks=pytest.mark.skip)])
            class Test1(object):
                def test_1(self, abc):
                    assert abc in {1, 2, 3}
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=3, failed=1, skipped=1)
        spans = self.pop_spans()

        assert len(spans) == 5

    def test_skip(self):
        """Test parametrize case."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.skip(reason="decorator")
            def test_decorator():
                pass

            def test_body():
                pytest.skip("body")
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(skipped=2)
        spans = self.pop_spans()

        assert len(spans) == 2
        assert spans[0].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[0].get_tag(test.SKIP_REASON) == "decorator"
        assert spans[1].get_tag(test.STATUS) == test.Status.SKIP.value
        assert spans[1].get_tag(test.SKIP_REASON) == "body"

    def test_xfail_fails(self):
        """Test xfail (expected failure) which fails, should be marked as pass."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail")
            def test_should_fail():
                assert 0
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        # pytest records xfail as skipped
        rec.assertoutcome(skipped=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XFAIL.value
        assert spans[0].get_tag(test.XFAIL_REASON) == "test should fail"

    def test_xpass_not_strict(self):
        """Test xpass (unexpected passing) with strict=False, should be marked as pass."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail")
            def test_should_fail():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XPASS.value
        assert spans[0].get_tag(test.XFAIL_REASON) == "test should fail"

    def test_xpass_strict(self):
        """Test xpass (unexpected passing) with strict=True, should be marked as fail."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.xfail(reason="test should fail", strict=True)
            def test_should_fail():
                pass
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(failed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag(test.STATUS) == test.Status.FAIL.value
        assert spans[0].get_tag(test.RESULT) == test.Status.XPASS.value
        # Note: xpass (strict=True) does not mark the reason with result.wasxfail but into result.longrepr,
        # however this provides the entire traceback/error into longrepr.
        assert "test should fail" in spans[0].get_tag(test.XFAIL_REASON)

    def test_tags(self):
        """Test ddspan tags."""
        py_file = self.testdir.makepyfile(
            """
            import pytest

            @pytest.mark.dd_tags(mark="dd_tags")
            def test_fixture(ddspan):
                assert ddspan is not None
                ddspan.set_tag("world", "hello")
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].get_tag("world") == "hello"
        assert spans[0].get_tag("mark") == "dd_tags"
        assert spans[0].get_tag(test.STATUS) == test.Status.PASS.value

    def test_default_service_name(self):
        """Test default service name."""
        py_file = self.testdir.makepyfile(
            """
            def test_service(ddspan):
                assert True
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.inline_run("--ddtrace", file_name)
        rec.assertoutcome(passed=1)
        spans = self.pop_spans()

        assert len(spans) == 1
        assert spans[0].service == "pytest"
        assert spans[0].name == "pytest.test"

    def test_dd_service_name(self):
        """Test integration service name."""
        self.monkeypatch.setenv("DD_SERVICE", "mysvc")
        if "DD_PYTEST_SERVICE" in os.environ:
            self.monkeypatch.delenv("DD_PYTEST_SERVICE")

        py_file = self.testdir.makepyfile(
            """
            import os

            def test_service(ddspan):
                assert 'mysvc' == os.getenv('DD_SERVICE')
                assert os.getenv('DD_PYTEST_SERVICE') is None
                assert 'mysvc' == ddspan.service
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        assert 0 == rec.ret

    def test_dd_pytest_service_name(self):
        """Test integration service name."""
        self.monkeypatch.setenv("DD_SERVICE", "mysvc")
        self.monkeypatch.setenv("DD_PYTEST_SERVICE", "pymysvc")
        self.monkeypatch.setenv("DD_PYTEST_OPERATION_NAME", "mytest")

        py_file = self.testdir.makepyfile(
            """
            import os

            def test_service(ddspan):
                assert 'mysvc' == os.getenv('DD_SERVICE')
                assert 'pymysvc' == os.getenv('DD_PYTEST_SERVICE')
                assert 'pymysvc' == ddspan.service
                assert 'mytest' == ddspan.name
        """
        )
        file_name = os.path.basename(py_file.strpath)
        rec = self.subprocess_run("--ddtrace", file_name)
        assert 0 == rec.ret



import pytest

import ddtrace

from ...constants import SPAN_KIND
from ...ext import SpanTypes
from ...ext import ci
from ...ext import test
from ...pin import Pin
from ..trace_utils import int_service
from .constants import FRAMEWORK
from .constants import HELP_MSG
from .constants import KIND


def is_enabled(config):
    """Check if the ddtrace plugin is enabled."""
    return config.getoption("ddtrace") or config.getini("ddtrace")


def _extract_span(item):
    """Extract span from `pytest.Item` instance."""
    return getattr(item, "_datadog_span", None)


def _store_span(item, span):
    """Store span at `pytest.Item` instance."""
    setattr(item, "_datadog_span", span)


PATCH_ALL_HELP_MSG = "Call ddtrace.patch_all before running tests."


def pytest_addoption(parser):
    """Add ddtrace options."""
    group = parser.getgroup("ddtrace")

    group._addoption(
        "--ddtrace",
        action="store_true",
        dest="ddtrace",
        default=False,
        help=HELP_MSG,
    )

    group._addoption(
        "--ddtrace-patch-all",
        action="store_true",
        dest="ddtrace-patch-all",
        default=False,
        help=PATCH_ALL_HELP_MSG,
    )

    parser.addini("ddtrace", HELP_MSG, type="bool")
    parser.addini("ddtrace-patch-all", PATCH_ALL_HELP_MSG, type="bool")


def pytest_configure(config):
    config.addinivalue_line("markers", "dd_tags(**kwargs): add tags to current span")

    if is_enabled(config):
        Pin(tags=ci.tags(), _config=ddtrace.config.pytest).onto(config)


def pytest_sessionfinish(session, exitstatus):
    """Flush open tracer."""
    pin = Pin.get_from(session.config)
    if pin is not None:
        pin.tracer.shutdown()


@pytest.fixture(scope="function")
def ddspan(request):
    pin = Pin.get_from(request.config)
    if pin:
        return _extract_span(request.node)


@pytest.fixture(scope="session", autouse=True)
def patch_all(request):
    if request.config.getoption("ddtrace-patch-all") or request.config.getini("ddtrace-patch-all"):
        ddtrace.patch_all()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_protocol(item, nextitem):
    pin = Pin.get_from(item.config)
    if pin is None:
        yield
        return

    with pin.tracer.trace(
        ddtrace.config.pytest.operation_name,
        service=int_service(pin, ddtrace.config.pytest),
        resource=item.nodeid,
        span_type=SpanTypes.TEST.value,
    ) as span:
        span.set_tags(pin.tags)
        span.set_tag(SPAN_KIND, KIND)
        span.set_tag(test.FRAMEWORK, FRAMEWORK)
        span.set_tag(test.NAME, item.name)
        span.set_tag(test.SUITE, item.module.__name__)
        span.set_tag(test.TYPE, SpanTypes.TEST.value)

        markers = [marker.kwargs for marker in item.iter_markers(name="dd_tags")]
        for tags in markers:
            span.set_tags(tags)
        _store_span(item, span)

        yield


def _extract_reason(call):
    if call.excinfo is not None:
        return call.excinfo.value


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Store outcome for tracing."""
    outcome = yield

    span = _extract_span(item)
    if span is None:
        return

    called_without_status = call.when == "call" and span.get_tag(test.STATUS) is None
    failed_setup = call.when == "setup" and call.excinfo is not None
    if not called_without_status and not failed_setup:
        return

    try:
        result = outcome.get_result()

        if hasattr(result, "wasxfail") or "xfail" in result.keywords:
            if result.skipped:
                # XFail tests that fail are recorded skipped by pytest
                span.set_tag(test.RESULT, test.Status.XFAIL.value)
                span.set_tag(test.XFAIL_REASON, result.wasxfail)
            else:
                span.set_tag(test.RESULT, test.Status.XPASS.value)
                if result.passed:
                    # XPass (strict=False) are recorded passed by pytest
                    span.set_tag(test.XFAIL_REASON, result.wasxfail)
                else:
                    # XPass (strict=True) are recorded failed by pytest, longrepr contains reason
                    span.set_tag(test.XFAIL_REASON, result.longrepr)

        if result.skipped:
            if hasattr(result, "wasxfail"):
                # XFail tests that fail are recorded skipped by pytest, should be passed instead
                span.set_tag(test.STATUS, test.Status.PASS.value)
            else:
                span.set_tag(test.STATUS, test.Status.SKIP.value)
            reason = _extract_reason(call)
            if reason is not None:
                span.set_tag(test.SKIP_REASON, reason)
        elif result.passed:
            span.set_tag(test.STATUS, test.Status.PASS.value)
        else:
            raise RuntimeWarning(result)
    except Exception:
        span.set_traceback()
        span.set_tag(test.STATUS, test.Status.FAIL.value)



from falcon import testing

import ddtrace
from ddtrace.contrib.falcon.patch import FALCON_VERSION
from tests.utils import TracerTestCase

from .app import get_app
from .test_suite import FalconTestCase


class AutoPatchTestCase(TracerTestCase, testing.TestCase, FalconTestCase):

    # Added because falcon 1.3 and 1.4 test clients (falcon.testing.client.TestClient) expect this property to be
    # defined. It would be initialized in the constructor, but we call it here like in 'TestClient.__init__(self, None)'
    # because falcon 1.0.x does not have such module and would fail. Once we stop supporting falcon 1.0.x then we can
    # use the cleaner __init__ invocation
    _default_headers = None

    def setUp(self):
        super(AutoPatchTestCase, self).setUp()

        self._service = "my-falcon"

        # Since most integrations do `from ddtrace import tracer` we cannot update do `ddtrace.tracer = self.tracer`
        self.original_writer = ddtrace.tracer.writer
        ddtrace.tracer.configure(writer=self.tracer.writer)
        self.tracer = ddtrace.tracer

        # build a test app without adding a tracer middleware;
        # reconfigure the global tracer since the autopatch mode
        # uses it
        self.api = get_app(tracer=None)

        if FALCON_VERSION >= (2, 0, 0):
            self.client = testing.TestClient(self.api)
        else:
            self.client = self

    def tearDown(self):
        super(AutoPatchTestCase, self).tearDown()

        ddtrace.tracer.writer = self.original_writer



import pytest

from ddtrace.internal import service


def test_service_status():
    s = service.Service()
    assert s.status == service.ServiceStatus.STOPPED
    s.start()
    assert s.status == service.ServiceStatus.RUNNING
    s.stop()
    assert s.status == service.ServiceStatus.STOPPED
    s.start()
    assert s.status == service.ServiceStatus.RUNNING
    s.stop()
    assert s.status == service.ServiceStatus.STOPPED


def test_service_status_on_fail():
    class ServiceFail(service.Service):
        def _start(self):
            raise RuntimeError

    s = ServiceFail()
    with pytest.raises(RuntimeError):
        s.start()
    assert s.status == service.ServiceStatus.STOPPED



import pytest

from ddtrace.internal.buffer import BufferFull
from ddtrace.internal.buffer import BufferItemTooLarge
from ddtrace.internal.buffer import TraceBuffer


def test_buffer_put_get():
    buf = TraceBuffer(max_size=10, max_item_size=3)
    buf.put([1, 2, 3])
    assert len(buf) == 1
    assert buf.size == 3

    buf.put([3, 2])
    assert len(buf) == 2
    assert buf.size == 5

    items = buf.get()
    assert len(buf) == 0
    assert buf.size == 0
    assert items == [[1, 2, 3], [3, 2]]

    buf.put("123")
    buf.put("45")
    buf.put("67")
    buf.put("89")
    buf.put("0")
    assert buf.size == 10
    assert len(buf) == 5
    assert buf.get() == ["123", "45", "67", "89", "0"]
    assert buf.size == 0


def test_buffer_size_limit():
    buf = TraceBuffer(max_size=10, max_item_size=3)
    buf.put([3])
    for i in range(9):
        buf.put([0])

    with pytest.raises(BufferFull):
        buf.put([1])

    with pytest.raises(BufferFull):
        buf.put([1])


def test_buffer_item_size_limit():
    buf = TraceBuffer(max_size=10, max_item_size=3)

    with pytest.raises(BufferItemTooLarge):
        buf.put([i for i in range(10000)])

    with pytest.raises(BufferItemTooLarge):
        buf.put([1, 2, 3, 4])

    buf.put([1, 2, 3])
    assert buf.size == 3


def test_buffer_many_items():
    buf = TraceBuffer(max_size=10000, max_item_size=50)
    for _ in range(10000):
        buf.put("1")
    assert buf.size == 10000
    assert len(buf) == 10000

    items = buf.get()
    assert len(items) == 10000
    for i in items:
        assert i == "1"

    for _ in range(200):
        buf.put("1" * 50)
    assert buf.size == 10000
    assert len(buf) == 200

    with pytest.raises(BufferFull):
        buf.put("1")
    items = buf.get()
    assert len(items) == 200
    for i in items:
        assert i == "1" * 50



from typing import Any
from typing import Set
from typing import TYPE_CHECKING
from typing import Tuple

from ddtrace.contrib.trace_utils import flatten_dict


if TYPE_CHECKING:
    from ddtrace.span import Span


EXCLUDED_ENDPOINT = {"kms", "sts"}
EXCLUDED_ENDPOINT_TAGS = {
    "s3": {"params.Body"},
    "firehose": {"params.Records"},
}


def truncate_arg_value(value, max_len=1024):
    # type: (Any, int) -> Any
    """Truncate values which are bytes and greater than `max_len`.
    Useful for parameters like 'Body' in `put_object` operations.
    """
    if isinstance(value, bytes) and len(value) > max_len:
        return b"..."

    return value


def add_span_arg_tags(
    span,  # type: Span
    endpoint_name,  # type: str
    args,  # type: Tuple[Any]
    args_names,  # type: Tuple[str]
    args_traced,  # type: Set[str]
):
    # type: (...) -> None
    if endpoint_name not in EXCLUDED_ENDPOINT:
        tags = dict((name, value) for (name, value) in zip(args_names, args) if name in args_traced)
        flat_tags = flatten_dict(tags, exclude=EXCLUDED_ENDPOINT_TAGS.get(endpoint_name))
        span.set_tags({k: truncate_arg_value(v) for k, v in flat_tags.items()})


REGION = "aws.region"
AGENT = "aws.agent"
OPERATION = "aws.operation"



from functools import wraps
import sys

import ddtrace

from .constants import FUTURE_SPAN_KEY
from .constants import PARENT_SPAN_KEY
from .stack_context import TracerStackContext


def _finish_span(future):
    """
    Finish the span if it's attached to the given ``Future`` object.
    This method is a Tornado callback used to close a decorated function
    executed as a coroutine or as a synchronous function in another thread.
    """
    span = getattr(future, FUTURE_SPAN_KEY, None)

    if span:
        # `tornado.concurrent.Future` in PY3 tornado>=4.0,<5 has `exc_info`
        if callable(getattr(future, "exc_info", None)):
            # retrieve the exception from the coroutine object
            exc_info = future.exc_info()
            if exc_info:
                span.set_exc_info(*exc_info)
        elif callable(getattr(future, "exception", None)):
            # in tornado>=4.0,<5 with PY2 `concurrent.futures._base.Future`
            # `exception_info()` returns `(exception, traceback)` but
            # `exception()` only returns the first element in the tuple
            if callable(getattr(future, "exception_info", None)):
                exc, exc_tb = future.exception_info()
                if exc and exc_tb:
                    exc_type = type(exc)
                    span.set_exc_info(exc_type, exc, exc_tb)
            # in tornado>=5 with PY3, `tornado.concurrent.Future` is alias to
            # `asyncio.Future` in PY3 `exc_info` not available, instead use
            # exception method
            else:
                exc = future.exception()
                if exc:
                    # we expect exception object to have a traceback attached
                    if hasattr(exc, "__traceback__"):
                        exc_type = type(exc)
                        exc_tb = getattr(exc, "__traceback__", None)
                        span.set_exc_info(exc_type, exc, exc_tb)
                    # if all else fails use currently handled exception for
                    # current thread
                    else:
                        span.set_exc_info(*sys.exc_info())

        span.finish()


def _run_on_executor(run_on_executor, _, params, kw_params):
    """
    Wrap the `run_on_executor` function so that when a function is executed
    in a different thread, we pass the current parent Span to the intermediate
    function that will execute the original call. The original function
    is then executed within a `TracerStackContext` so that `tracer.trace()`
    can be used as usual, both with empty or existing `Context`.
    """

    def pass_context_decorator(fn):
        """
        Decorator that is used to wrap the original `run_on_executor_decorator`
        so that we can pass the current active context before the `executor.submit`
        is called. In this case we get the `parent_span` reference and we pass
        that reference to `fn` reference. Because in the outer wrapper we replace
        the original call with our `traced_wrapper`, we're sure that the `parent_span`
        is passed to our intermediate function and not to the user function.
        """

        @wraps(fn)
        def wrapper(*args, **kwargs):
            # from the current context, retrive the active span
            current_ctx = ddtrace.tracer.get_call_context()
            parent_span = getattr(current_ctx, "_current_span", None)

            # pass the current parent span in the Future call so that
            # it can be retrieved later
            kwargs.update({PARENT_SPAN_KEY: parent_span})
            return fn(*args, **kwargs)

        return wrapper

    # we expect exceptions here if the `run_on_executor` is called with
    # wrong arguments; in that case we should not do anything because
    # the exception must not be handled here
    decorator = run_on_executor(*params, **kw_params)

    # `run_on_executor` can be called with arguments; in this case we
    # return an inner decorator that holds the real function that should be
    # called
    if decorator.__module__ == "tornado.concurrent":

        def run_on_executor_decorator(deco_fn):
            def inner_traced_wrapper(*args, **kwargs):
                # retrieve the parent span from the function kwargs
                parent_span = kwargs.pop(PARENT_SPAN_KEY, None)
                return run_executor_stack_context(deco_fn, args, kwargs, parent_span)

            return pass_context_decorator(decorator(inner_traced_wrapper))

        return run_on_executor_decorator

    # return our wrapper function that executes an intermediate function to
    # trace the real execution in a different thread
    def traced_wrapper(*args, **kwargs):
        # retrieve the parent span from the function kwargs
        parent_span = kwargs.pop(PARENT_SPAN_KEY, None)
        return run_executor_stack_context(params[0], args, kwargs, parent_span)

    return pass_context_decorator(run_on_executor(traced_wrapper))


def run_executor_stack_context(fn, args, kwargs, parent_span):
    """
    This intermediate function is always executed in a newly created thread. Here
    using a `TracerStackContext` is legit because this function doesn't interfere
    with the main thread loop. `StackContext` states are thread-local and retrieving
    the context here will always bring to an empty `Context`.
    """
    with TracerStackContext():
        ctx = ddtrace.tracer.get_call_context()
        ctx._current_span = parent_span
        return fn(*args, **kwargs)


def wrap_executor(tracer, fn, args, kwargs, span_name, service=None, resource=None, span_type=None):
    """
    Wrap executor function used to change the default behavior of
    ``Tracer.wrap()`` method. A decorated Tornado function can be
    a regular function or a coroutine; if a coroutine is decorated, a
    span is attached to the returned ``Future`` and a callback is set
    so that it will close the span when the ``Future`` is done.
    """
    span = tracer.trace(span_name, service=service, resource=resource, span_type=span_type)

    # catch standard exceptions raised in synchronous executions
    try:
        future = fn(*args, **kwargs)

        # duck-typing: if it has `add_done_callback` it's a Future
        # object whatever is the underlying implementation
        if callable(getattr(future, "add_done_callback", None)):
            setattr(future, FUTURE_SPAN_KEY, span)
            future.add_done_callback(_finish_span)
        else:
            # we don't have a future so the `future` variable
            # holds the result of the function
            span.finish()
    except Exception:
        span.set_traceback()
        span.finish()
        raise

    return future



"""
testing config.
"""

import os


# default config for backing services
# NOTE: defaults may be duplicated in the .env file; update both or
# simply write down a function that parses the .env file

ELASTICSEARCH_CONFIG = {
    'port': int(os.getenv('TEST_ELASTICSEARCH_PORT', 9200)),
}

CASSANDRA_CONFIG = {
    'port': int(os.getenv('TEST_CASSANDRA_PORT', 9042)),
}

CONSUL_CONFIG = {
    'host': '127.0.0.1',
    'port': int(os.getenv('TEST_CONSUL_PORT', 8500)),
}

# Use host=127.0.0.1 since local docker testing breaks with localhost

POSTGRES_CONFIG = {
    'host': '127.0.0.1',
    'port': int(os.getenv('TEST_POSTGRES_PORT', 5432)),
    'user': os.getenv('TEST_POSTGRES_USER', 'postgres'),
    'password': os.getenv('TEST_POSTGRES_PASSWORD', 'postgres'),
    'dbname': os.getenv('TEST_POSTGRES_DB', 'postgres'),
}

MYSQL_CONFIG = {
    'host': '127.0.0.1',
    'port': int(os.getenv('TEST_MYSQL_PORT', 3306)),
    'user': os.getenv('TEST_MYSQL_USER', 'test'),
    'password': os.getenv('TEST_MYSQL_PASSWORD', 'test'),
    'database': os.getenv('TEST_MYSQL_DATABASE', 'test'),
}

REDIS_CONFIG = {
    'port': int(os.getenv('TEST_REDIS_PORT', 6379)),
}

REDISCLUSTER_CONFIG = {
    'host': '127.0.0.1',
    'ports': os.getenv('TEST_REDISCLUSTER_PORTS', '7000,7001,7002,7003,7004,7005'),
}

MONGO_CONFIG = {
    'port': int(os.getenv('TEST_MONGO_PORT', 27017)),
}

MEMCACHED_CONFIG = {
    'host': os.getenv('TEST_MEMCACHED_HOST', '127.0.0.1'),
    'port': int(os.getenv('TEST_MEMCACHED_PORT', 11211)),
}

VERTICA_CONFIG = {
    'host': os.getenv('TEST_VERTICA_HOST', '127.0.0.1'),
    'port': os.getenv('TEST_VERTICA_PORT', 5433),
    'user': os.getenv('TEST_VERTICA_USER', 'dbadmin'),
    'password': os.getenv('TEST_VERTICA_PASSWORD', 'abc123'),
    'database': os.getenv('TEST_VERTICA_DATABASE', 'docker'),
}

RABBITMQ_CONFIG = {
    'host': os.getenv('TEST_RABBITMQ_HOST', '127.0.0.1'),
    'user': os.getenv('TEST_RABBITMQ_USER', 'guest'),
    'password': os.getenv('TEST_RABBITMQ_PASSWORD', 'guest'),
    'port': int(os.getenv('TEST_RABBITMQ_PORT', 5672)),
}



import gevent

from ...context import Context
from ...provider import BaseContextProvider


# Greenlet attribute used to set/get the Context instance
CONTEXT_ATTR = "__datadog_context"


class GeventContextProvider(BaseContextProvider):
    """
    Context provider that retrieves all contexts for the current asynchronous
    execution. It must be used in asynchronous programming that relies
    in the ``gevent`` library. Framework instrumentation that uses the
    gevent WSGI server (or gevent in general), can use this provider.
    """

    def _get_current_context(self):
        """Helper to get the current context from the current greenlet"""
        current_g = gevent.getcurrent()
        if current_g is not None:
            return getattr(current_g, CONTEXT_ATTR, None)
        return None

    def _has_active_context(self):
        """Helper to determine if we have a currently active context"""
        return self._get_current_context() is not None

    def activate(self, context):
        """Sets the scoped ``Context`` for the current running ``Greenlet``."""
        current_g = gevent.getcurrent()
        if current_g is not None:
            setattr(current_g, CONTEXT_ATTR, context)
            return context

    def active(self):
        """
        Returns the scoped ``Context`` for this execution flow. The ``Context``
        uses the ``Greenlet`` class as a carrier, and everytime a greenlet
        is created it receives the "parent" context.
        """
        ctx = self._get_current_context()
        if ctx is not None:
            # return the active Context for this greenlet (if any)
            return ctx

        # the Greenlet doesn't have a Context so it's created and attached
        # even to the main greenlet. This is required in Distributed Tracing
        # when a new arbitrary Context is provided.
        current_g = gevent.getcurrent()
        if current_g:
            ctx = Context()
            setattr(current_g, CONTEXT_ATTR, ctx)
            return ctx



"""
An API to provide after_in_child fork hooks across all Pythons.
"""
import logging
import os
import typing


log = logging.getLogger(__name__)


_registry = []  # type: typing.List[typing.Callable[[], None]]


def ddtrace_after_in_child():
    # type: () -> None
    global _registry

    # DEV: we make a copy of the registry to prevent hook execution from
    # introducing new hooks, potentially causing an infinite loop.
    for hook in list(_registry):
        try:
            hook()
        except Exception:
            # Mimic the behaviour of Python's fork hooks.
            log.exception("Exception ignored in forksafe hook %r", hook)


def register(after_in_child):
    # type: (typing.Callable[[], None]) -> typing.Callable[[], None]
    """Register a function to be called after fork in the child process."""
    _registry.append(after_in_child)
    return after_in_child


def unregister(after_in_child):
    # type: (typing.Callable[[], None]) -> None
    """Unregister a function to be called after fork in the child process.

    Raises `ValueError` if the function was not registered.
    """
    _registry.remove(after_in_child)


if hasattr(os, "register_at_fork"):
    os.register_at_fork(after_in_child=ddtrace_after_in_child)
else:
    import threading

    _threading_after_fork = threading._after_fork  # type: ignore[attr-defined]

    def _after_fork():
        # type: () -> None
        _threading_after_fork()
        ddtrace_after_in_child()

    threading._after_fork = _after_fork  # type: ignore[attr-defined]



import threading
from typing import Optional
from typing import TYPE_CHECKING

from .constants import ORIGIN_KEY
from .constants import SAMPLING_PRIORITY_KEY
from .internal.logger import get_logger
from .utils.deprecation import deprecated
from .utils.formats import asbool
from .utils.formats import get_env


if TYPE_CHECKING:
    from ddtrace import Span

log = get_logger(__name__)


class Context(object):
    """
    Context is used to keep track of a hierarchy of spans for the current
    execution flow. During each logical execution, the same ``Context`` is
    used to represent a single logical trace, even if the trace is built
    asynchronously.

    A single code execution may use multiple ``Context`` if part of the execution
    must not be related to the current tracing. As example, a delayed job may
    compose a standalone trace instead of being related to the same trace that
    generates the job itself. On the other hand, if it's part of the same
    ``Context``, it will be related to the original trace.

    This data structure is thread-safe.
    """

    _partial_flush_enabled = asbool(get_env("tracer", "partial_flush_enabled", default=False))
    _partial_flush_min_spans = int(get_env("tracer", "partial_flush_min_spans", default=500))  # type: ignore[arg-type]

    def __init__(
        self,
        trace_id=None,  # type: Optional[int]
        span_id=None,  # type: Optional[int]
        sampling_priority=None,  # type: Optional[int]
        dd_origin=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        """
        Initialize a new thread-safe ``Context``.

        :param int trace_id: trace_id of parent span
        :param int span_id: span_id of parent span
        """
        self._current_span = None  # type: Optional[Span]
        self._lock = threading.RLock()

        self._parent_trace_id = trace_id
        self._parent_span_id = span_id
        self._sampling_priority = sampling_priority
        self._local_root_span = None  # type: Optional[Span]
        self.dd_origin = dd_origin

    @property
    def trace_id(self):
        """Return current context trace_id."""
        with self._lock:
            return self._parent_trace_id

    @property
    def span_id(self):
        """Return current context span_id."""
        with self._lock:
            return self._parent_span_id

    @property
    def sampling_priority(self):
        """Return current context sampling priority."""
        with self._lock:
            return self._sampling_priority

    @sampling_priority.setter
    def sampling_priority(self, value):
        # type: (int) -> None
        """Set sampling priority."""
        with self._lock:
            self._sampling_priority = value

    def _clone(self):
        # type: () -> Context
        with self._lock:
            new_ctx = Context(
                trace_id=self._parent_trace_id,
                span_id=self._parent_span_id,
                sampling_priority=self._sampling_priority,
            )
            new_ctx._current_span = self._current_span
            return new_ctx

    @deprecated("Cloning contexts will no longer be required in 0.50", version="0.50")
    def clone(self):
        # type: () -> Context
        """
        Partially clones the current context.
        It copies everything EXCEPT the registered and finished spans.
        """
        return self._clone()

    def _get_current_root_span(self):
        # type: () -> Optional[Span]
        with self._lock:
            return self._local_root_span

    @deprecated("Please use tracer.current_root_span() instead", version="0.50")
    def get_current_root_span(self):
        # type: () -> Optional[Span]
        """
        Return the root span of the context or None if it does not exist.
        """
        return self._get_current_root_span()

    def _get_current_span(self):
        # type: () -> Optional[Span]
        with self._lock:
            return self._current_span

    @deprecated("Please use tracer.current_span() instead", version="0.50")
    def get_current_span(self):
        # type: () -> Optional[Span]
        """
        Return the last active span that corresponds to the last inserted
        item in the trace list. This cannot be considered as the current active
        span in asynchronous environments, because some spans can be closed
        earlier while child spans still need to finish their traced execution.
        """
        return self._get_current_span()

    def _set_current_span(self, span):
        # type: (Optional[Span]) -> None
        """
        Set current span internally.

        Non-safe if not used with a lock. For internal Context usage only.
        """
        self._current_span = span
        if span:
            self._parent_trace_id = span.trace_id
            self._parent_span_id = span.span_id
        else:
            self._parent_span_id = None

    def _add_span(self, span):
        # type: (Span) -> None
        with self._lock:
            # Assume the first span added to the context is the local root
            if self._local_root_span is None:
                self._local_root_span = span
            self._set_current_span(span)
            span._context = self

    @deprecated("Context will no longer support active span management in a later version.", version="0.50")
    def add_span(self, span):
        # type: (Span) -> None
        """Activates span in the context."""
        return self._add_span(span)

    def _close_span(self, span):
        # type: (Span) -> None
        with self._lock:
            if span == self._local_root_span:
                if self.dd_origin is not None:
                    span.meta[ORIGIN_KEY] = self.dd_origin
                if self._sampling_priority is not None:
                    span.metrics[SAMPLING_PRIORITY_KEY] = self._sampling_priority

            # If a parent exists to the closing span and it is unfinished, then
            # activate it next.
            if span._parent and not span._parent.finished:
                self._set_current_span(span._parent)
            # Else if the span is the local root of this context, then clear the
            # context so the next trace can be started.
            elif span == self._local_root_span:
                self._set_current_span(span._parent)
                self._local_root_span = None
                self._parent_trace_id = None
                self._parent_span_id = None
                self._sampling_priority = None
            # Else the span that is closing is closing after its parent.
            # This is most likely an error. To ensure future traces are not
            # affected clear out the context and set the current span to
            # ``None``.
            else:
                log.debug(
                    "span %r closing after its parent %r, this is an error when not using async", span, span._parent
                )
                self._set_current_span(None)
                self._local_root_span = None
                self._parent_trace_id = None
                self._parent_span_id = None
                self._sampling_priority = None

    @deprecated(message="Context will no longer support active span management in a later version.", version="0.50")
    def close_span(self, span):
        # type: (Span) -> None
        """Updates the context after a span has finished.

        The next active span becomes `span`'s parent.
        """
        return self._close_span(span)



from hypothesis import given

from ddtrace import tracer
from ddtrace.contrib.trace_utils import flatten_dict
from ddtrace.ext.aws import add_span_arg_tags
from tests.tracer.test_trace_utils import nested_dicts


def test_flatten_dict_hypothesis(benchmark):
    @given(nested_dicts)
    def _wrapper(d):
        flatten_dict(d)

    benchmark(_wrapper)


NESTED_DICT = {
    "http": {
        "url": "https://www.example.com/v2/service/123456/get?token=abf545e1ff55acd55ca&view=list#somefragment",
        "query_string": "?token=abf545e1ff55acd55ca&view=list",
        "request_headers": {
            "content_type": "plain/text",
            "http-x-my-header": "foobar",
        },
    },
    "params": {
        "Record": {
            "http": {
                "url": "https://www.example.com/v2/service/123456/get?token=abf545e1ff55acd55ca&view=list#somefragment",
                "query_string": "?token=abf545e1ff55acd55ca&view=list",
                "request_headers": {
                    "content_type": "plain/text",
                    "http-x-my-header": "foobar",
                },
            }
        },
        "Body": {
            "http": {
                "url": "https://www.example.com/v2/service/123456/get?token=abf545e1ff55acd55ca&view=list#somefragment",
                "query_string": "?token=abf545e1ff55acd55ca&view=list",
                "request_headers": {
                    "content_type": "plain/text",
                    "http-x-my-header": "foobar",
                },
            }
        },
    },
}


def test_flatten_dict_full(benchmark):
    benchmark(flatten_dict, NESTED_DICT)


def test_flatten_dict_exclude(benchmark):
    benchmark(flatten_dict, NESTED_DICT, exclude={"params.Record"})


def test_add_span_arg_tags(benchmark):
    def _(tracer):
        with tracer.trace("foo") as span:
            add_span_arg_tags(
                span, "s3", (NESTED_DICT["http"], NESTED_DICT["params"]), ("http", "params"), ("http", "params")
            )

    benchmark(_, tracer)



from collections import namedtuple


TAG_NAMES = [
    "RESOURCE_NAME",
    "SAMPLING_PRIORITY",
    "SERVICE_NAME",
    "SPAN_TYPE",
    "TARGET_HOST",
    "TARGET_PORT",
]

TagNames = namedtuple("TagNames", TAG_NAMES)

Tags = TagNames(
    RESOURCE_NAME="resource.name",
    SAMPLING_PRIORITY="sampling.priority",
    SERVICE_NAME="service.name",
    TARGET_HOST="out.host",
    TARGET_PORT="out.port",
    SPAN_TYPE="span.type",
)



from collections import deque
import threading
from typing import Deque
from typing import List

import attr


class BufferFull(Exception):
    pass


class BufferItemTooLarge(Exception):
    pass


@attr.s
class TraceBuffer(object):
    """A thread-safe buffer for collecting encoded trace payloads to be sent to
    a Datadog Agent.

    :param max_size: The maximum size (in bytes) of the buffer.
    :param max_item_size: The maximum size of any item in the buffer. It is necessary
        to have an item limit as traces cannot be divided across trace payloads.
    """

    max_size = attr.ib(type=int)
    max_item_size = attr.ib(type=int)
    _size = attr.ib(init=False, type=int, default=0, repr=True)
    _lock = attr.ib(init=False, factory=threading.Lock, repr=False)
    _buffer = attr.ib(init=False, factory=deque, repr=False, type=Deque[bytes])

    def __len__(self):
        return len(self._buffer)

    @property
    def size(self):
        # type: () -> int
        """Return the size in bytes of the trace buffer."""
        with self._lock:
            return self._size

    def _clear(self):
        # type: () -> None
        self._buffer.clear()
        self._size = 0

    def put(self, item):
        # type: (bytes) -> None
        """Put an item in the buffer.

        The item should be an encoded trace (list of spans).
        """
        item_len = len(item)
        if item_len > self.max_item_size or item_len > self.max_size:
            raise BufferItemTooLarge()

        with self._lock:
            if self._size + item_len <= self.max_size:
                self._buffer.append(item)
                self._size += item_len
            else:
                raise BufferFull()

    def get(self):
        # type: () -> List[bytes]
        """Return the entire buffer.

        The buffer is cleared in the process.
        """
        with self._lock:
            try:
                return list(self._buffer)
            finally:
                self._clear()



"""
The Django__ integration traces requests, views, template renderers, database
and cache calls in a Django application.


Enable Django tracing automatically via ``ddtrace-run``::

    ddtrace-run python manage.py runserver


Django tracing can also be enabled manually::

    from ddtrace import patch_all
    patch_all()


To have Django capture the tracer logs, ensure the ``LOGGING`` variable in
``settings.py`` looks similar to::

    LOGGING = {
        'loggers': {
            'ddtrace': {
                'handlers': ['console'],
                'level': 'WARNING',
            },
        },
    }


Configuration
~~~~~~~~~~~~~
.. py:data:: ddtrace.config.django['distributed_tracing_enabled']

   Whether or not to parse distributed tracing headers from requests received by your Django app.

   Default: ``True``

.. py:data:: ddtrace.config.django['service_name']

   The service name reported for your Django app.

   Can also be configured via the ``DD_SERVICE`` environment variable.

   Default: ``'django'``

.. py:data:: ddtrace.config.django['cache_service_name']

   The service name reported for your Django app cache layer.

   Can also be configured via the ``DD_DJANGO_CACHE_SERVICE_NAME`` environment variable.

   Default: ``'django'``

.. py:data:: ddtrace.config.django['database_service_name']

   A string reported as the service name of the Django app database layer.

   Can also be configured via the ``DD_DJANGO_DATABASE_SERVICE_NAME`` environment variable.

   Takes precedence over database_service_name_prefix.

   Default: ``''``

.. py:data:: ddtrace.config.django['database_service_name_prefix']

   A string to be prepended to the service name reported for your Django app database layer.

   Can also be configured via the ``DD_DJANGO_DATABASE_SERVICE_NAME_PREFIX`` environment variable.

   The database service name is the name of the database appended with 'db'. Has a lower precedence than database_service_name.

   Default: ``''``

.. py:data:: ddtrace.config.django["trace_fetch_methods"]

   Whether or not to trace fetch methods.

   Can also configured via the ``DD_DJANGO_TRACE_FETCH_METHODS`` environment variable.

   Default: ``False``

.. py:data:: ddtrace.config.django['instrument_middleware']

   Whether or not to instrument middleware.

   Can also be enabled with the ``DD_DJANGO_INSTRUMENT_MIDDLEWARE`` environment variable.

   Default: ``True``

.. py:data:: ddtrace.config.django['instrument_databases']

   Whether or not to instrument databases.

   Default: ``True``

.. py:data:: ddtrace.config.django['instrument_caches']

   Whether or not to instrument caches.

   Default: ``True``

.. py:data:: ddtrace.config.django['trace_query_string']

   Whether or not to include the query string as a tag.

   Default: ``False``

.. py:data:: ddtrace.config.django['include_user_name']

   Whether or not to include the authenticated user's username as a tag on the root request span.

   Default: ``True``

.. py:data:: ddtrace.config.django['use_handler_resource_format']

   Whether or not to use the resource format `"{method} {handler}"`. Can also be
   enabled with the ``DD_DJANGO_USE_HANDLER_RESOURCE_FORMAT`` environment
   variable.

   The default resource format for Django >= 2.2.0 is otherwise `"{method} {urlpattern}"`.

   Default: ``False``

.. py:data:: ddtrace.config.django['use_legacy_resource_format']

   Whether or not to use the legacy resource format `"{handler}"`. Can also be
   enabled with the ``DD_DJANGO_USE_LEGACY_RESOURCE_FORMAT`` environment
   variable.

   The default resource format for Django >= 2.2.0 is otherwise `"{method} {urlpattern}"`.

   Default: ``False``

Example::

    from ddtrace import config

    # Enable distributed tracing
    config.django['distributed_tracing_enabled'] = True

    # Override service name
    config.django['service_name'] = 'custom-service-name'


Migration from ddtrace<=0.33.0
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
The Django integration provides automatic migration from enabling tracing using
a middleware to the method consistent with our integrations. Application
developers are encouraged to convert their configuration of the tracer to the
latter.

1. Remove ``'ddtrace.contrib.django'`` from ``INSTALLED_APPS`` in
   ``settings.py``.

2. Replace ``DATADOG_TRACE`` configuration in ``settings.py`` according to the
   table below.

3. Remove ``TraceMiddleware`` or ``TraceExceptionMiddleware`` if used in
   ``settings.py``.

4. Enable Django tracing automatically via `ddtrace-run`` or manually by
   adding ``ddtrace.patch_all()`` to ``settings.py``.

5. Set environment variable ``DD_DJANGO_USE_LEGACY_RESOURCE_FORMAT`` or
   ``ddtrace.config.django['use_legacy_resource_format']`` to continue using the
   legacy resource format `"{handler}"` rather than the new default resource
   format `"{method} {urlpattern}"`.

The mapping from old configuration settings to new ones.

+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``DATADOG_TRACE``           | Configuration                                                                                                           |
+=============================+=========================================================================================================================+
| ``AGENT_HOSTNAME``          | ``DD_AGENT_HOST`` environment variable or ``tracer.configure(hostname=)``                                               |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``AGENT_PORT``              | ``DD_TRACE_AGENT_PORT`` environment variable or ``tracer.configure(port=)``                                             |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``AUTO_INSTRUMENT``         | N/A Instrumentation is automatic                                                                                        |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``INSTRUMENT_CACHE``        | ``config.django['instrument_caches']``                                                                                  |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``INSTRUMENT_DATABASE``     | ``config.django['instrument_databases']``                                                                               |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``INSTRUMENT_TEMPLATE``     | N/A Instrumentation is automatic                                                                                        |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``DEFAULT_DATABASE_PREFIX`` | ``config.django['database_service_name_prefix']``                                                                       |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``DEFAULT_SERVICE``         | ``DD_SERVICE_NAME`` environment variable or ``config.django['service_name']``                                           |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``DEFAULT_CACHE_SERVICE``   | ``config.django['cache_service_name']``                                                                                 |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``ENABLED``                 | ``tracer.configure(enabled=)``                                                                                          |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``DISTRIBUTED_TRACING``     | ``config.django['distributed_tracing_enabled']``                                                                        |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``TRACE_QUERY_STRING``      | ``config.django['trace_query_string']``                                                                                 |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``TAGS``                    | ``DD_TAGS`` environment variable or ``tracer.set_tags()``                                                               |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+
| ``TRACER``                  | N/A - if a particular tracer is required for the Django integration use ``Pin.override(Pin.get_from(django), tracer=)`` |
+-----------------------------+-------------------------------------------------------------------------------------------------------------------------+

Examples
--------
Before::

   # settings.py
   INSTALLED_APPS = [
       # your Django apps...
       'ddtrace.contrib.django',
   ]

   DATADOG_TRACE = {
       'AGENT_HOSTNAME': 'localhost',
       'AGENT_PORT': 8126,
       'AUTO_INSTRUMENT': True,
       'INSTRUMENT_CACHE': True,
       'INSTRUMENT_DATABASE': True,
       'INSTRUMENT_TEMPLATE': True,
       'DEFAULT_SERVICE': 'my-django-app',
       'DEFAULT_CACHE_SERVICE': 'my-cache',
       'DEFAULT_DATABASE_PREFIX': 'my-',
       'ENABLED': True,
       'DISTRIBUTED_TRACING': True,
       'TRACE_QUERY_STRING': None,
       'TAGS': {'env': 'production'},
       'TRACER': 'my.custom.tracer',
   }

After::

   # settings.py
   INSTALLED_APPS = [
       # your Django apps...
   ]

   from ddtrace import config, tracer
   tracer.configure(hostname='localhost', port=8126, enabled=True)
   config.django['service_name'] = 'my-django-app'
   config.django['cache_service_name'] = 'my-cache'
   config.django['database_service_name_prefix'] = 'my-'
   config.django['instrument_databases'] = True
   config.django['instrument_caches'] = True
   config.django['trace_query_string'] = True
   tracer.set_tags({'env': 'production'})

   import my.custom.tracer
   from ddtrace import Pin, patch_all
   import django
   patch_all()
   Pin.override(Pin.get_from(django), tracer=my.custom.tracer)

:ref:`Headers tracing <http-headers-tracing>` is supported for this integration.

.. __: https://www.djangoproject.com/
"""  # noqa: E501
from ...utils.importlib import require_modules


required_modules = ["django"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from . import patch as _patch
        from .middleware import TraceMiddleware
        from .patch import patch
        from .patch import unpatch

        __all__ = ["patch", "unpatch", "TraceMiddleware", "_patch"]


# define the Django app configuration
default_app_config = "ddtrace.contrib.django.apps.TracerConfig"



import os
import uuid


__all__ = [
    "get_runtime_id",
]


def _generate_runtime_id():
    return uuid.uuid4().hex


_RUNTIME_ID = _generate_runtime_id()


if hasattr(os, "register_at_fork"):

    def _set_runtime_id():
        global _RUNTIME_ID
        _RUNTIME_ID = _generate_runtime_id()

    os.register_at_fork(after_in_child=_set_runtime_id)

    def get_runtime_id():
        return _RUNTIME_ID


else:
    # Non-POSIX systems or Python < 3.7
    import threading

    _RUNTIME_PID = os.getpid()

    def _set_runtime_id():
        global _RUNTIME_ID, _RUNTIME_PID
        _RUNTIME_ID = _generate_runtime_id()
        _RUNTIME_PID = os.getpid()

    _RUNTIME_LOCK = threading.Lock()

    def get_runtime_id():
        with _RUNTIME_LOCK:
            pid = os.getpid()
            if pid != _RUNTIME_PID:
                _set_runtime_id()
            return _RUNTIME_ID


get_runtime_id.__doc__ = """Return a unique string identifier for this runtime.

Do not store this identifier as it can change when, e.g., the process forks."""



import os

from ddtrace.internal import forksafe


def test_forksafe():
    state = []

    @forksafe.register
    def after_in_child():
        state.append(1)

    def my_func():
        return state

    pid = os.fork()

    if pid == 0:
        # child
        assert my_func() == [1]
        os._exit(12)
    else:
        assert my_func() == []

    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_registry():
    state = []

    @forksafe.register
    def after_in_child_1():
        state.append(1)

    def f1():
        pass

    @forksafe.register
    def after_in_child_2():
        state.append(2)

    def f2():
        pass

    @forksafe.register
    def after_in_child_3():
        state.append(3)

    def f3():
        pass

    forksafe.ddtrace_after_in_child()
    assert state == [1, 2, 3]


def test_duplicates():
    state = []

    @forksafe.register
    def hook():
        state.append(1)

    def f1():
        return state

    def f2():
        return state

    def f3():
        return state

    pid = os.fork()

    if pid == 0:
        # child
        assert f1() == f2() == f3() == [1]
        os._exit(12)
    else:
        assert f1() == f2() == f3() == []

    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_method_usage():
    class A:
        def __init__(self):
            self.state = 0
            forksafe.register(self.after_fork)

        def after_fork(self):
            self.state = 1

        def method(self):
            return self.state

    a = A()
    pid = os.fork()
    if pid == 0:
        # child
        assert a.method() == 1
        os._exit(12)
    else:
        assert a.method() == 0

    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12


def test_hook_exception():
    state = []

    @forksafe.register
    def after_in_child():
        raise ValueError

    @forksafe.register
    def state_append_1():
        forksafe.register(state.append(1))

    @forksafe.register
    def state_append_3():
        forksafe.register(state.append(3))

    def f1():
        return state

    def f2():
        return state

    def f3():
        return state

    pid = os.fork()
    if pid == 0:
        # child
        assert f1() == f2() == f3() == [1, 3]
        os._exit(12)
    else:
        assert f1() == f2() == f3() == []

    _, status = os.waitpid(pid, 0)
    exit_code = os.WEXITSTATUS(status)
    assert exit_code == 12



import msgpack
from msgpack.fallback import Packer
import pytest

from ddtrace.internal.encoding import MsgpackEncoder
from ddtrace.internal.encoding import _EncoderBase
from tests.tracer.test_encoders import RefMsgpackEncoder
from tests.tracer.test_encoders import gen_trace


msgpack_encoder = RefMsgpackEncoder()
trace_encoder = MsgpackEncoder()


class PPMsgpackEncoder(_EncoderBase):
    content_type = "application/msgpack"

    @staticmethod
    def encode(obj):
        return Packer().pack(obj)

    @staticmethod
    def decode(data):
        return msgpack.unpackb(data, raw=True)


trace_large = gen_trace(nspans=1000)
trace_small = gen_trace(nspans=50, key_size=10, ntags=5, nmetrics=4)


@pytest.mark.benchmark(group="encoding.join_encoded", min_time=0.005)
def test_join_encoded(benchmark):
    benchmark(
        msgpack_encoder.join_encoded,
        [msgpack_encoder.encode_trace(trace_large), msgpack_encoder.encode_trace(trace_small)],
    )


@pytest.mark.benchmark(group="encoding", min_time=0.005)
def test_encode_1000_span_trace(benchmark):
    benchmark(msgpack_encoder.encode_trace, trace_large)


@pytest.mark.benchmark(group="encoding.small", min_time=0.005)
def test_encode_trace_small(benchmark):
    benchmark(msgpack_encoder.encode_trace, trace_small)


@pytest.mark.benchmark(group="encoding.small.multi", min_time=0.005)
def test_encode_trace_small_multi(benchmark):
    benchmark(msgpack_encoder.encode_traces, [trace_small for _ in range(50)])


@pytest.mark.benchmark(group="encoding", min_time=0.005)
def test_encode_1000_span_trace_fallback(benchmark):
    encoder = PPMsgpackEncoder()
    benchmark(encoder.encode_trace, trace_large)


@pytest.mark.benchmark(group="encoding", min_time=0.005)
def test_encode_1000_span_trace_custom(benchmark):
    benchmark(trace_encoder.encode_trace, trace_large)


@pytest.mark.benchmark(group="encoding.small", min_time=0.005)
def test_encode_trace_small_custom(benchmark):
    benchmark(trace_encoder.encode_trace, trace_small)


@pytest.mark.benchmark(group="encoding.small.multi", min_time=0.005)
def test_encode_trace_small_multi_custom(benchmark):
    benchmark(trace_encoder.encode_traces, [trace_small for _ in range(50)])


@pytest.mark.benchmark(group="encoding.join_encoded", min_time=0.005)
def test_join_encoded_custom(benchmark):
    benchmark(
        trace_encoder.join_encoded, [trace_encoder.encode_trace(trace_large), trace_encoder.encode_trace(trace_small)]
    )



# 3p
import rediscluster

# project
from ddtrace import config
from ddtrace.vendor import wrapt

from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import redis as redisx
from ...pin import Pin
from ...utils.wrappers import unwrap
from ..redis.patch import traced_execute_command
from ..redis.patch import traced_pipeline
from ..redis.util import format_command_args


# DEV: In `2.0.0` `__version__` is a string and `VERSION` is a tuple,
#      but in `1.x.x` `__version__` is a tuple annd `VERSION` does not exist
REDISCLUSTER_VERSION = getattr(rediscluster, 'VERSION', rediscluster.__version__)


def patch():
    """Patch the instrumented methods
    """
    if getattr(rediscluster, '_datadog_patch', False):
        return
    setattr(rediscluster, '_datadog_patch', True)

    _w = wrapt.wrap_function_wrapper
    if REDISCLUSTER_VERSION >= (2, 0, 0):
        _w('rediscluster', 'client.RedisCluster.execute_command', traced_execute_command)
        _w('rediscluster', 'client.RedisCluster.pipeline', traced_pipeline)
        _w('rediscluster', 'pipeline.ClusterPipeline.execute', traced_execute_pipeline)
        Pin(service=redisx.DEFAULT_SERVICE, app=redisx.APP).onto(rediscluster.RedisCluster)
    else:
        _w('rediscluster', 'StrictRedisCluster.execute_command', traced_execute_command)
        _w('rediscluster', 'StrictRedisCluster.pipeline', traced_pipeline)
        _w('rediscluster', 'StrictClusterPipeline.execute', traced_execute_pipeline)
        Pin(service=redisx.DEFAULT_SERVICE, app=redisx.APP).onto(rediscluster.StrictRedisCluster)


def unpatch():
    if getattr(rediscluster, '_datadog_patch', False):
        setattr(rediscluster, '_datadog_patch', False)

        if REDISCLUSTER_VERSION >= (2, 0, 0):
            unwrap(rediscluster.client.RedisCluster, 'execute_command')
            unwrap(rediscluster.client.RedisCluster, 'pipeline')
            unwrap(rediscluster.pipeline.ClusterPipeline, 'execute')
        else:
            unwrap(rediscluster.StrictRedisCluster, 'execute_command')
            unwrap(rediscluster.StrictRedisCluster, 'pipeline')
            unwrap(rediscluster.StrictClusterPipeline, 'execute')


#
# tracing functions
#

def traced_execute_pipeline(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return func(*args, **kwargs)

    cmds = [format_command_args(c.args) for c in instance.command_stack]
    resource = '\n'.join(cmds)
    tracer = pin.tracer
    with tracer.trace(redisx.CMD, resource=resource, service=pin.service, span_type=SpanTypes.REDIS) as s:
        s.set_tag(SPAN_MEASURED_KEY)
        s.set_tag(redisx.RAWCMD, resource)
        s.set_metric(redisx.PIPELINE_LEN, len(instance.command_stack))

        # set analytics sample rate if enabled
        s.set_tag(
            ANALYTICS_SAMPLE_RATE_KEY,
            config.rediscluster.get_analytics_sample_rate()
        )

        return func(*args, **kwargs)



import asyncio

from ...context import Context
from ...provider import DefaultContextProvider


# Task attribute used to set/get the Context instance
CONTEXT_ATTR = "__datadog_context"


class AsyncioContextProvider(DefaultContextProvider):
    """
    Context provider that retrieves all contexts for the current asyncio
    execution. It must be used in asynchronous programming that relies
    in the built-in ``asyncio`` library. Framework instrumentation that
    is built on top of the ``asyncio`` library, can use this provider.

    This Context Provider inherits from ``DefaultContextProvider`` because
    it uses a thread-local storage when the ``Context`` is propagated to
    a different thread, than the one that is running the async loop.
    """

    def activate(self, context, loop=None):
        """Sets the scoped ``Context`` for the current running ``Task``."""
        loop = self._get_loop(loop)
        if not loop:
            super(AsyncioContextProvider, self).activate(context)
            return context

        # the current unit of work (if tasks are used)
        task = asyncio.Task.current_task(loop=loop)
        setattr(task, CONTEXT_ATTR, context)
        return context

    def _get_loop(self, loop=None):
        """Helper to try and resolve the current loop"""
        try:
            return loop or asyncio.get_event_loop()
        except RuntimeError:
            # Detects if a loop is available in the current thread;
            # DEV: This happens when a new thread is created from the out that is running the async loop
            # DEV: It's possible that a different Executor is handling a different Thread that
            #      works with blocking code. In that case, we fallback to a thread-local Context.
            pass
        return None

    def _has_active_context(self, loop=None):
        """Helper to determine if we have a currently active context"""
        loop = self._get_loop(loop=loop)
        if loop is None:
            return super(AsyncioContextProvider, self)._has_active_context()

        # the current unit of work (if tasks are used)
        task = asyncio.Task.current_task(loop=loop)
        if task is None:
            return False

        ctx = getattr(task, CONTEXT_ATTR, None)
        return ctx is not None

    def active(self, loop=None):
        """
        Returns the scoped Context for this execution flow. The ``Context`` uses
        the current task as a carrier so if a single task is used for the entire application,
        the context must be handled separately.
        """
        loop = self._get_loop(loop=loop)
        if not loop:
            return super(AsyncioContextProvider, self).active()

        # the current unit of work (if tasks are used)
        task = asyncio.Task.current_task(loop=loop)
        if task is None:
            # providing a detached Context from the current Task, may lead to
            # wrong traces. This defensive behavior grants that a trace can
            # still be built without raising exceptions
            return Context()

        ctx = getattr(task, CONTEXT_ATTR, None)
        if ctx is not None:
            # return the active Context for this task (if any)
            return ctx

        # create a new Context using the Task as a Context carrier
        ctx = Context()
        setattr(task, CONTEXT_ATTR, ctx)
        return ctx
